from yated.ptypes import Point
import re
from . import config


def normalize(row, tabsize, tab_count):
    row = row.replace('\n', '')
    tab_count[0] += row.count('\t')
    row = row.replace('\t', ' ' * tabsize)
    return row


class Document:
    def __init__(self, filename=''):
        self.clear()
        self.filename = filename
        self.tabs_replaced = False
        self.text = ['']
        self.valid = False
        self.undos = []
        self.undoing = False
        self.modified = False

    def clear(self):
        self.filename = ''
        self.text = ['']
        self.valid = False
        self.undos = []
        self.undoing = False
        self.modified = False

    def load(self, filename):
        try:
            self.tabs_replaced = False
            f = open(filename, 'r')
            self.text = f.readlines()
            f.close()
            if len(self.text) == 0:
                self.text = ['']
            else:
                if self.text[-1].endswith('\n'):
                    self.text.append('')
            tabsize = config.getint('tabsize')
            tab_count = [0]
            self.text = [normalize(row, tabsize, tab_count) for row in self.text]
            if tab_count[0] > 0:
                self.tabs_replaced = True
            self.filename = filename
            self.invalidate()
        except IOError:
            pass
            # message_box('Failed to open {}'.format(filename))

    def save_as(self, path):
        self.filename = path
        return self.save()

    def save(self):
        if len(self.filename) == 0:
            return False
        f = open(self.filename, 'w')
        f.write('\n'.join(self.text))
        f.close()
        self.modified = False
        return True

    def rows_count(self):
        return len(self.text)

    def get_row(self, index):
        if index < 0 or index >= self.rows_count():
            return ''
        return self.text[index]

    def find_in_row(self, cursor, text, case, regex, word):
        if cursor.y < 0 or cursor.y >= self.rows_count():
            return -1
        row = self.get_row(cursor.y)
        if not case:
            row = row.lower()
            text = text.lower()
        if word:
            text = r'(^|\W){}($|\W)'.format(text)
            regex = True
        if regex:
            pat = re.compile(text)
            m = pat.search(row, cursor.x + 1)
            if m:
                return 1 + m.start()
            return -1
        else:
            return row.find(text, cursor.x)

    def join_next_row(self, row_index):
        if 0 <= row_index < (self.rows_count() - 1):
            row = self.get_row(row_index)
            next_row = self.get_row(row_index + 1)
            del self.text[row_index + 1]
            self.text[row_index] = row + next_row
            if not self.undoing:
                self.undos.append([self.new_line, Point(len(row), row_index)])
            self.modified = True
        return Point(0, 0)

    def insert_block(self, text, cursor):
        if cursor.y < 0 or cursor.y >= self.rows_count():
            return False
        row = self.get_row(cursor.y)
        if cursor.x <= len(row):
            if not self.undoing:
                self.undos.append([self.delete_block, cursor.y, cursor.x, cursor.x + len(text)])
            row = row[0:cursor.x] + text + row[cursor.x:]
            self.text[cursor.y] = row
        return Point(len(text), 0)

    def delete_block(self, row_index, x0, x1):
        if 0 <= row_index < self.rows_count():
            row = self.get_row(row_index)
            if x1 < x0:
                x1 = len(row)
            if not self.undoing:
                self.undos.append([self.insert_block, row[x0:x1], Point(x0, row_index)])
            self.text[row_index] = row[0:x0] + row[x1:]
            self.modified = True
        return Point(0, 0)

    def insert_row(self, row_index, text=''):
        if 0 <= row_index < self.rows_count():
            if not self.undoing:
                self.undos.append([self.delete_row, row_index])
            self.text.insert(row_index, text)
        return Point(0, 0)

    def delete_row(self, row_index):
        if 0 <= row_index < self.rows_count():
            if not self.undoing:
                self.undos.append([self.insert_row, row_index, self.get_row(row_index)])
            del self.text[row_index]
            self.modified = True
        return Point(0, 0)

    def word_right(self, cursor):
        if cursor.y < 0 or cursor.y >= self.rows_count():
            return cursor
        row = self.get_row(cursor.y)
        x = cursor.x
        space_found = False
        if x < len(row):
            while x < len(row):
                if not row[x].isalnum():
                    space_found = True
                if row[x].isalnum() and space_found:
                    break
                x = x + 1
        else:
            if cursor.y < (self.rows_count() - 1):
                return Point(0, cursor.y + 1)
        return Point(x, cursor.y)

    def word_left(self, cursor):
        if cursor.y < 0 or cursor.y >= self.rows_count():
            return cursor
        row = self.get_row(cursor.y)
        x = cursor.x
        if x == 0:
            if cursor.y > 0:
                return Point(len(self.get_row(cursor.y - 1)), cursor.y - 1)
        while x > 0:
            x -= 1
            if x > 0 and row[x].isalnum() and not row[x - 1].isalnum():
                break
        return Point(x, cursor.y)

    def set_cursor(self, cursor):
        if cursor.y < 0:
            return Point(0, 0)
        if cursor.y >= self.rows_count():
            return Point(0, self.rows_count() - 1)
        row = self.get_row(cursor.y)
        x = cursor.x
        if x < 0:
            x = 0
        if x > len(row):
            x = len(row)
        return Point(x, cursor.y)

    def delete_char(self, cursor):
        row = self.text[cursor.y]
        if cursor.x < len(row):
            if not self.undoing:
                self.undos.append([self.add_char, row[cursor.x], Point(cursor), True])
            row = row[0:cursor.x] + row[cursor.x + 1:]
            self.text[cursor.y] = row
            self.invalidate()
        else:
            self.join_next_row(cursor.y)
        self.modified = True
        return Point(0, 0)

    def add_char(self, c, cursor, insert):
        row = self.text[cursor.y]
        if not insert and cursor.x < len(row):
            self.delete_char(cursor)
        row = row[0:cursor.x] + c + row[cursor.x:]
        if not self.undoing:
            self.undos.append([self.delete_char, Point(cursor)])
        self.text[cursor.y] = row
        self.invalidate()
        self.modified = True
        return Point(1, 0)

    def add_text(self, text, cursor, insert):
        cx = cursor.x
        res = Point(0, 0)
        if cursor.x < 0 or cursor.y < 0 or cursor.y >= self.rows_count():
            return res
        for c in text:
            if ord(c) == 10:
                self.new_line(cursor)
                cursor = Point(0, cursor.y + 1)
                res = Point(-cx, res.y + 1)
            else:
                self.add_char(c, cursor, insert)
                cursor = cursor + Point(1, 0)
                res += (1, 0)
        self.modified = True
        return res

    def new_line(self, cursor):
        row = self.get_row(cursor.y)
        cur = [row[0:cursor.x], row[cursor.x:]]
        self.text = self.text[0:cursor.y] + cur + self.text[cursor.y + 1:]
        if not self.undoing:
            self.undos.append([self.join_next_row, cursor.y])
        self.modified = True
        return Point(-cursor.x, 1)

    def invalidate(self):
        self.valid = False

    def validate(self):
        self.valid = True

    def start_compound(self):
        if not self.undoing:
            self.undos.append('{')

    def stop_compound(self):
        if not self.undoing:
            self.undos.append('}')

    def undo(self):
        depth = 0
        self.undoing = True
        res = Point(0, 0)
        while len(self.undos) > 0:
            cmd = self.undos[-1]
            del self.undos[-1]
            if isinstance(cmd, str):
                if cmd == '{':
                    depth += 1
                if cmd == '}':
                    depth -= 1
            else:
                f = cmd[0]
                res += f(*cmd[1:])
            if depth == 0:
                break
        self.undoing = False
        return res



from yated.ptypes import Point, Rect
from . import config
import curses
import os


class Dialog(object):
    def __init__(self, width, height):
        self.width = width
        self.height = height
        self.rect = None

    def draw(self, app):
        cx = int(app.width / 2)
        cy = int(app.height / 2)
        y = int(cy - self.height / 2)
        x = int(cx - self.width / 2)
        self.rect = Rect(x, y, x + self.width, y + self.height)
        app.draw_frame(self.rect, 4)
        app.fill_rect(Rect(self.rect).inflate(-1), ' ', 4)


class MessageBox(Dialog):
    def __init__(self, prompt):
        super(MessageBox, self).__init__(60, 8)
        self.prompt = prompt
        self.actions = {}

    def add_key(self, key, action):
        self.actions[key.lower()] = action
        self.actions[key.upper()] = action

    def process_key(self, key):
        if key in self.actions:
            return self.actions.get(key)
        return None

    def draw(self, app):
        super(MessageBox, self).draw(app)
        app.move(Point(self.rect.tl.x + 4, self.rect.tl.y + 4))
        app.write(self.prompt, 4)


class AboutDialog(Dialog):
    def __init__(self):
        super(AboutDialog, self).__init__(40, 20)

    def draw(self, app):
        super(AboutDialog, self).draw(app)
        pos = self.rect.tl + Point(4, 3)
        lines = ('Yet Another Text EDitor',
                 '',
                 'A terminal based text editor, ',
                 'inspired by a wish to have',
                 'an editor with default visual',
                 'and key bindings that resemble',
                 'typical GUI based editors.',
                 '',
                 'https://github.com/amirgeva/yated'
                 )
        for line in lines:
            app.move(pos)
            app.write(line, 4)
            pos += (0, 1)

    def process_key(self, key):
        return True


class FileDialog(Dialog):
    def __init__(self, action):
        super(FileDialog, self).__init__(60, 20)
        self.dir = os.getcwd()
        self.browse_mode = True
        self.edit_text = ''
        self.edit_position = None
        self.ofs = 0
        self.cur = 0
        self.items = []
        self.fill_items()
        self.action = action

    def fill_items(self):
        files = os.listdir(self.dir)
        for i in range(0, len(files)):
            name = files[i]
            path = os.path.join(self.dir, name)
            if os.path.isdir(path):
                name = name + '/'
            files[i] = name
        files.insert(0, '../')
        self.items = files

    def draw(self, app):
        super(FileDialog, self).draw(app)
        pos = self.rect.tl + Point(1, 1)
        app.move(pos)
        app.write(self.dir + '/', 1)
        self.edit_position = pos + (len(self.dir) + 1, 0)
        pos += (2, 1)
        for i in range(0, 17):
            c = 4
            idx = i + self.ofs
            app.move(pos)
            if idx < len(self.items):
                attr = 0
                if self.cur == idx:
                    c = 1
                s = self.items[idx]
                if len(s) > 50:
                    s = s[0:50]
                if len(s) < 50:
                    s = s + ' ' * (50 - len(s))
                app.write(s, c, attr)
            else:
                app.write(' ' * 50, c)
            pos += (0, 1)
        if not self.browse_mode:
            app.move(self.edit_position)
            app.write(self.edit_text, 1)

    def process_key(self, key):
        if self.browse_mode:
            if key == chr(9):  # tab
                self.browse_mode = False
            if key == 'KEY_DOWN':
                self.cur = (self.cur + 1) % len(self.items)
            if key == 'KEY_UP':
                self.cur = (self.cur - 1) % len(self.items)
            if key == 'KEY_END':
                self.cur = len(self.items) - 1
            if key == 'KEY_HOME':
                self.cur = 0
            if key == 'KEY_PPAGE':
                self.cur = (self.cur - 17) % len(self.items)
            if key == 'KEY_NPAGE':
                self.cur = (self.cur + 17) % len(self.items)
            if key == chr(10):  # Enter
                name = self.items[self.cur]
                path = os.path.abspath(os.path.join(self.dir, name))
                if name.endswith('/'):
                    self.dir = path
                    self.fill_items()
                else:
                    return lambda: self.action(path)
            if self.cur < self.ofs or self.cur >= (self.ofs + 17):
                self.ofs = self.cur - 8
                if self.ofs < 0:
                    self.ofs = 0
        else:
            if key == chr(9):  # tab
                self.browse_mode = True
            if len(key) == 1 and 32 < ord(key) < 128:
                self.edit_text += key
            if key == 'KEY_BACKSPACE' and len(self.edit_text) > 0:
                self.edit_text = self.edit_text[0:-1]
            if key == chr(10):
                path = os.path.abspath(os.path.join(self.dir, self.edit_text))
                return lambda: self.action(path)
        return None


class FormWidget(object):
    def __init__(self, pos):
        self.text = ''
        self.pos = pos
        self.tabstop = False


class FormCheckbox(FormWidget):
    def __init__(self, pos, state=False):
        super(FormCheckbox, self).__init__(pos)
        self.tabstop = True
        self.state = state

    def details(self):
        return self.state

    def draw(self, app, tl):
        app.move(self.pos + tl - Point(1, 0))
        mark = 'X' if self.state else ' '
        app.write('[{}]'.format(mark), 4)
        app.move(self.pos + tl)

    def process_key(self, key):
        if key == ' ':
            self.state = not self.state


class FormLabel(FormWidget):
    def __init__(self, pos, text):
        super(FormLabel, self).__init__(pos)
        self.text = text
        self.color = 4
        self.pad = 0

    def draw(self, app, tl):
        app.move(self.pos + tl)
        app.write(self.text, self.color)
        if self.pad > len(self.text):
            app.write(' ' * (self.pad - len(self.text)), self.color)


class FormEdit(FormLabel):
    def __init__(self, pos, text, maxlen):
        super(FormEdit, self).__init__(pos, text)
        self.tabstop = True
        self.color = 1
        self.maxlen = maxlen
        self.pad = maxlen

    def details(self):
        return self.text

    def draw(self, app, tl):
        super(FormEdit, self).draw(app, tl)
        app.move(self.pos + tl + Point(len(self.text), 0))

    def process_key(self, key):
        if len(key) == 1 and 32 <= ord(key[0]) < 128:
            self.text = self.text + key
        if key == 'KEY_BACKSPACE' and len(self.text) > 0:
            self.text = self.text[0:-1]


class FormDialog(Dialog):
    def __init__(self, w, h):
        super(FormDialog, self).__init__(w, h)
        self.widgets = []
        self.named_widgets = {}
        self.cur = None

    def details(self, vals=None):
        if vals is None:
            vals = []
        res = {}
        for val in vals:
            res[val[0]] = val[1]
        for name in self.named_widgets:
            res[name] = self.named_widgets.get(name).details()
        return res

    def add_widget(self, name, widget):
        if isinstance(widget, FormWidget):
            self.widgets.append(widget)
            if name:
                self.named_widgets[name] = widget
            if self.cur is None and widget.tabstop:
                self.cur = len(self.widgets) - 1

    def draw(self, app):
        super(FormDialog, self).draw(app)
        for i in range(0, len(self.widgets)):
            widget = self.widgets[i]
            if self.cur != i:
                widget.draw(app, self.rect.tl)
        if self.cur is not None:
            self.widgets[self.cur].draw(app, self.rect.tl)

    def process_key(self, key):
        if key == chr(9):  # tab
            while True:
                self.cur = (self.cur + 1) % len(self.widgets)
                if self.widgets[self.cur].tabstop:
                    break
        if key == 'KEY_BTAB':  # back-tab
            while True:
                self.cur = (self.cur - 1) % len(self.widgets)
                if self.widgets[self.cur].tabstop:
                    break
        if self.cur is not None:
            self.widgets[self.cur].process_key(key)
        return None


class FindReplaceDialog(FormDialog):
    def __init__(self, find_action, replace_action):
        super(FindReplaceDialog, self).__init__(60, 20)
        self.find_action = find_action
        self.replace_action = replace_action
        self.add_widget('', FormLabel(Point(2, 2), 'Find:'))
        self.add_widget('find', FormEdit(Point(12, 2), config.get('find_text'), 40))
        self.add_widget('', FormLabel(Point(2, 3), 'Replace:'))
        self.add_widget('replace', FormEdit(Point(12, 3), config.get('find_replace'), 40))
        self.add_widget('case', FormCheckbox(Point(4, 5), config.getbool('find_case')))
        self.add_widget('', FormLabel(Point(8, 5), 'Case Sensitive'))
        self.add_widget('regex', FormCheckbox(Point(4, 6), config.getbool('find_regex')))
        self.add_widget('', FormLabel(Point(8, 6), 'Regular Expressions'))
        self.add_widget('word', FormCheckbox(Point(4, 7), config.getbool('find_word')))
        self.add_widget('', FormLabel(Point(8, 7), 'Whole Word'))
        #
        self.add_widget('', FormLabel(Point(2, 16), 'Enter to find,  Ctrl+R to replace,  Ctrl+A to replace all'))

    def process_key(self, key):
        props = [('all', False), ('first', True)]
        if key == chr(18):  # Ctrl+R
            return lambda: self.replace_action(self.details(props))
        if key == chr(1):  # Ctrl+A
            props[0] = 'all', True
            return lambda: self.replace_action(self.details(props))
        if key == chr(10):  # Enter
            return lambda: self.find_action(self.details(props))
        return super(FindReplaceDialog, self).process_key(key)


class ColorConfigDialog(Dialog):
    def __init__(self):
        super(ColorConfigDialog, self).__init__(60, 15)
        self.cur = 1

    def draw(self, app):
        super(ColorConfigDialog, self).draw(app)
        pos = self.rect.tl + Point(2, 2)
        for i in range(1, 8):
            app.move(pos)
            attr = 0
            c = 4
            if i == self.cur:
                c = 1
            app.write(f'Color {i}  ', c, attr)
            app.write('ABC', i)
            pos += (0, 1)
        pos += Point(8, 1)
        app.move(pos)
        app.write('Use left/right pgup/pgdn to change colors', 4)
        app.move(pos + (0, 1))
        app.write('Esc when done', 4)

    def process_key(self, key):
        if key == 'KEY_DOWN':
            self.cur = 1 + (self.cur % 7)
        if key == 'KEY_UP':
            self.cur = 1 + ((self.cur - 2) % 7)

        fg = config.getint('fg{}'.format(self.cur))
        bg = config.getint('bg{}'.format(self.cur))
        if key == 'KEY_LEFT':
            fg = (fg - 1) % 8
        if key == 'KEY_RIGHT':
            fg = (fg + 1) % 8
        if key == 'KEY_PPAGE':
            bg = (bg - 1) % 8
        if key == 'KEY_NPAGE':
            bg = (bg + 1) % 8
        config.set('fg{}'.format(self.cur), fg)
        config.set('bg{}'.format(self.cur), bg)
        curses.init_pair(self.cur, fg, bg)
        return None


class EditorOptionsDialog(FormDialog):
    def __init__(self, action):
        super(EditorOptionsDialog, self).__init__(60, 20)
        self.action = action
        self.add_widget('', FormLabel(Point(2, 2), 'Tab Size:'))
        self.add_widget('tabsize', FormEdit(Point(16, 2), config.get('tabsize', '4'), 2))
        self.add_widget('', FormLabel(Point(2, 16), 'Enter to save,  Esc to cancel'))

    def process_key(self, key):
        super(EditorOptionsDialog, self).process_key(key)
        if key == chr(10):
            return lambda: self.action(self.details())
        return None



#!/usr/bin/env python3


class Point(object):
    def __init__(self, *args):
        if len(args) == 2:
            self.x = args[0]
            self.y = args[1]
        elif len(args) == 1:
            if isinstance(args[0], Point):
                self.x = args[0].x
                self.y = args[0].y
            elif isinstance(args[0], tuple):
                self.x = args[0][0]
                self.y = args[0][1]
            else:
                raise TypeError()

    def __str__(self):
        return '{},{}'.format(self.x, self.y)

    def __lt__(self, p):
        if not isinstance(p, Point):
            p = Point(p)
        if self.y < p.y:
            return True
        if self.y > p.y:
            return False
        return self.x < p.x

    def __le__(self, p):
        return self < p or self == p

    def __gt__(self, p):
        return not (self < p or self == p)

    def __ge__(self, p):
        return not (self < p)

    def __eq__(self, p):
        if not isinstance(p, Point):
            p = Point(p)
        return self.x == p.x and self.y == p.y

    def __ne__(self, p):
        return not (self == p)

    def __iadd__(self, p):
        if not isinstance(p, Point):
            p = Point(p)
        self.x += p.x
        self.y += p.y
        return self

    def __isub__(self, p):
        if not isinstance(p, Point):
            p = Point(p)
        self.x -= p.x
        self.y -= p.y
        return self

    def __add__(self, p):
        if not isinstance(p, Point):
            p = Point(p)
        return Point(self.x + p.x, self.y + p.y)

    def __sub__(self, p):
        if not isinstance(p, Point):
            p = Point(p)
        return Point(self.x - p.x, self.y - p.y)


class Rect(object):
    def __init__(self, *args):
        if len(args) == 2 and isinstance(args[0], Point) and isinstance(args[1], Point):
            self.tl = Point(args[0])
            self.br = Point(args[1])
        elif len(args) == 4:
            self.tl = Point(args[0], args[1])
            self.br = Point(args[2], args[3])
        elif len(args) == 1 and isinstance(args[0], Rect):
            self.tl = Point(args[0].tl)
            self.br = Point(args[0].br)
        else:
            raise TypeError()

    def normalized(self):
        return Rect(min(self.tl.x, self.br.x), min(self.tl.y, self.br.y),
                    max(self.tl.x, self.br.x), max(self.tl.y, self.br.y))

    def width(self):
        return self.br.x - self.tl.x

    def height(self):
        return self.br.y - self.tl.y

    def inflate(self, d):
        if isinstance(d, Point):
            self.tl -= d
            self.br += d
        else:
            self.tl -= Point(d, d)
            self.br += Point(d, d)
        return self

    def is_point_inside(self, p):
        if not isinstance(p, Point):
            p = Point(p)
        return self.br.x > p.x >= self.tl.x and self.br.y > p.y >= self.tl.y


def unit_test():
    p1 = Point(1, 2)
    p2 = Point((1, 2))
    p3 = Point(p1)
    print(str(p1))
    print(str(p2))
    print(str(p3))


if __name__ == '__main__':
    unit_test()



import curses
from yated.ptypes import Point


class MenuItem(object):
    def __init__(self, title, action=None):
        self.title = title
        self.action = action
        self.key = ''
        p = title.index('&')
        if 0 <= p < (len(title) - 1):
            self.key = title[p + 1]

    def activate(self):
        if self.action is not None:
            self.action()


class Menu(object):
    def __init__(self, title):
        self.title = title
        self.pos = Point(0, 0)
        self.key = ''
        self.cur = 0
        self.items = []
        self.width = 4

    def add_item(self, *args):  # title, action = None):
        if len(args) == 1 and isinstance(args[0], MenuItem):
            item = args[0]
        elif len(args) == 1 and isinstance(args[0], Menu):
            item = args[0]
        elif len(args) == 1 and isinstance(args[0], str):
            item = MenuItem(args[0])
        elif len(args) == 2:
            item = MenuItem(args[0], args[1])
        else:
            raise TypeError()
        self.items.append(item)
        self.width = max(self.width, 4 + len(item.title))

    def on_key(self, key):
        for item in self.items:
            if item.key == key:
                item.activate()
                return True
        return False

    def select_next(self):
        self.cur = (self.cur + 1) % len(self.items)

    def select_prev(self):
        self.cur = (self.cur - 1) % len(self.items)

    def activate_current(self):
        self.items[self.cur].activate()

    def draw(self, scr):
        color = 4
        pos = Point(self.pos)
        scr.move(pos)
        scr.write(curses.ACS_ULCORNER, color)
        for i in range(0, self.width - 2):
            scr.write(curses.ACS_HLINE, color)
        scr.write(curses.ACS_URCORNER, color)
        index = 0
        for item in self.items:
            color = 4
            pos += (0, 1)
            scr.move(pos)
            scr.write(curses.ACS_VLINE, color)
            scr.write(' ', color)
            n = 0
            if index == self.cur:
                color = 3
            char_color = color
            for c in item.title:
                if c == '&':
                    char_color = 5
                else:
                    attr = 0
                    scr.write(c, char_color, attr)
                    char_color = color
                    n = n + 1
            s = ' ' * (self.width - 4 - n)
            scr.write(s, color)
            color = 4
            scr.write(' ', color)
            scr.write(curses.ACS_VLINE, color)
            index += 1
        pos += (0, 1)
        scr.move(pos)
        scr.write(curses.ACS_LLCORNER, color)
        for i in range(0, self.width - 2):
            scr.write(curses.ACS_HLINE, color)
        scr.write(curses.ACS_LRCORNER, color)



#!/usr/bin/env python3
import sys
import atexit
import subprocess

impl = None


class qt_Clipboard:
    def __init__(self):
        from PyQt5 import QtWidgets
        self.app = QtWidgets.QApplication(sys.argv)
        self.clipboard = self.app.clipboard()

    def copy(self, text):
        self.clipboard.setText(text)

    def paste(self):
        return self.clipboard.text()

    def close(self):
        self.app.exit()


class xclip_Clipboard:
    def __init__(self):
        subprocess.check_output(['xclip', '-version'])

    def copy(self, text):
        p = subprocess.Popen(['xclip', '-i', '-selection', 'clipboard'], stdin=subprocess.PIPE)
        p.stdin.write(bytes(bytearray(text, 'utf-8')))
        p.stdin.close()

    def paste(self):
        try:
            return subprocess.check_output(['xclip', '-o', '-selection', 'clipboard']).decode("utf-8")
        except subprocess.CalledProcessError:
            return ''

    def close(self):
        pass


class internal_Clipboard:
    def __init__(self):
        self.text = ''

    def copy(self, text):
        self.text = text

    def paste(self):
        return self.text

    def close(self):
        pass


def cleanup():
    if impl is not None:
        impl.close()


atexit.register(cleanup)


def init_qt():
    global impl
    try:
        impl = qt_Clipboard()
    except ImportError:
        pass


def init_xclip():
    global impl
    try:
        impl = xclip_Clipboard()
    except Exception:
        pass


def copy(text):
    if impl is not None:
        impl.copy(text)


def paste():
    if impl is None:
        return ''
    return impl.paste()


if impl is None:
    init_xclip()
if impl is None:
    init_qt()
if impl is None:
    impl = internal_Clipboard()


def unit_test():
    copy("bla")
    if paste() != "bla":
        print("Failed")
    else:
        print("Success")


if __name__ == '__main__':
    unit_test()



#!/usr/bin/env python3
import curses
import sys
from . import utils
from . import config
from yated.ptypes import Point, Rect
from yated.doc import Document
from yated.view import View
from yated.menus import Menu


class Application(object):
    def __init__(self):
        self.scr = curses.initscr()
        self.keylog = None
        curses.noecho()
        curses.raw()
        self.scr.keypad(1)
        mx = self.scr.getmaxyx()
        self.width = mx[1]
        self.height = mx[0]
        self.rect = Rect(0, 0, self.width, self.height)
        self.menu_bar = Menu('')
        self.shortcuts = {}
        curses.start_color()
        curses.use_default_colors()
        curses.init_pair(1, config.getint('fg1', curses.COLOR_YELLOW),
                         config.getint('bg1', curses.COLOR_BLUE))
        curses.init_pair(2, config.getint('fg2', curses.COLOR_WHITE),
                         config.getint('bg2', curses.COLOR_GREEN))
        curses.init_pair(3, config.getint('fg3', curses.COLOR_BLACK),
                         config.getint('bg3', curses.COLOR_WHITE))
        curses.init_pair(4, config.getint('fg4', curses.COLOR_BLACK),
                         config.getint('bg4', curses.COLOR_CYAN))
        curses.init_pair(5, config.getint('fg5', curses.COLOR_YELLOW),
                         config.getint('bg5', curses.COLOR_BLUE))
        curses.init_pair(6, config.getint('fg6', curses.COLOR_YELLOW),
                         config.getint('bg6', curses.COLOR_BLUE))
        curses.init_pair(7, config.getint('fg7', curses.COLOR_WHITE),
                         config.getint('bg7', curses.COLOR_RED))
        sys.stdout.write('\033]12;yellow\007')

    def set_menu(self, bar):
        self.menu_bar = bar
        if len(bar.items) > 0:
            self.shortcuts['KEY_F(10)'] = bar.items[0]

    def move(self, pos):
        if not isinstance(pos, Point):
            pos = Point(pos)
        if self.rect.is_point_inside(pos):
            self.scr.move(pos.y, pos.x)
            return True
        return False

    def fill_rect(self, rect, c, clr):
        self.fill(rect.tl.x, rect.tl.y, rect.width(), rect.height(), c, clr)

    def fill(self, x0, y0, w, h, c, clr):
        for y in range(y0, y0 + h):
            self.move((x0, y))
            self.write(c * w, clr)
            # for x in range(x0, x0 + w):
            # self.scr.addch(c, curses.color_pair(clr))

    def write(self, text, clr, attr=curses.A_BOLD):
        clr = curses.color_pair(clr | (attr & 0x7FFF))
        if isinstance(text, str):
            for i in range(0, len(text)):
                c = text[i]
                self.scr.addstr(c, clr)
        else:
            self.scr.addch(text, clr)

    def draw_frame(self, rect, color):
        self.move(rect.tl)
        self.write(curses.ACS_ULCORNER, color)
        for i in range(0, rect.width() - 2):
            self.write(curses.ACS_HLINE, color)
        self.write(curses.ACS_URCORNER, color)
        for y in range(rect.tl.y + 1, rect.br.y):
            self.move(Point(rect.tl.x, y))
            self.write(curses.ACS_VLINE, color)
            self.move(Point(rect.br.x - 1, y))
            self.write(curses.ACS_VLINE, color)
        self.move(Point(rect.tl.x, rect.br.y - 1))
        self.write(curses.ACS_LLCORNER, color)
        for i in range(0, rect.width() - 2):
            self.write(curses.ACS_HLINE, color)
        self.write(curses.ACS_LRCORNER, color)

    def refresh(self):
        self.scr.refresh()

    def getkey(self):
        try:
            self.scr.nodelay(False)
            key = self.scr.getkey()
            if len(key) == 1 and ord(key[0]) == 27:
                self.scr.nodelay(True)
                key = "Alt+" + self.scr.getkey()
                self.scr.nodelay(False)
        except curses.error:
            key = 'ESC'
        if key == 'KEY_F(24)':  # For debugging purposes
            self.keylog = open('/tmp/key.log', 'w')
        if self.keylog is not None:
            self.keylog.write('{}   {}\n'.format(key, hex(ord(key[0]))))
            self.keylog.flush()
        return key

    def close(self):
        curses.nocbreak()
        self.scr.keypad(0)
        curses.echo()
        curses.endwin()
        self.scr = None

    def draw_menu(self):
        color = 4
        self.move((0, 0))
        self.write(' ' * self.width, color)
        self.move((1, 0))
        pos = Point(2, 1)
        for item in self.menu_bar.items:
            title = item.title
            item.pos = Point(pos)
            pos += Point(len(title) - title.count('&') + 3, 0)
            self.write('[', color)
            rev = False
            char_color = color
            for c in title:
                if c == '&':
                    char_color = color + 1
                    rev = True
                else:
                    attr = curses.A_BOLD
                    if rev:
                        rev = False
                        self.shortcuts['Alt+' + c.upper()] = item
                        self.shortcuts['Alt+' + c.lower()] = item
                    self.write(c, char_color, attr)
                    char_color = color
            self.write('] ', color)


def message_box(text):
    pass


def fill_menu(menu, desc):
    for item in desc:
        title = item[0]
        action = item[1]
        if isinstance(action, list):
            m = Menu(title)
            fill_menu(m, action)
            menu.add_item(m)
        else:
            menu.add_item(title, action)


def create_menu(view):
    desc = [('&File', [('&New     Ctrl+N', view.on_file_new),
                       ('&Open    Ctrl+O', view.on_file_open),
                       ('&Save    Ctrl+S', view.on_file_save),
                       ('Save &As', view.on_file_save_as),
                       ('&Exit    Ctrl+Q', view.on_file_exit)
                       ]),
            ('&Edit', [('&Copy          Ctrl+C', view.on_copy),
                       ('C&ut           Ctrl+X', view.on_cut),
                       ('&Paste         Ctrl+V', view.on_paste),
                       ('&Find          Ctrl+F', view.on_find_replace),
                       ('Find &Again        F3', view.on_find_again),
                       ('&Record Macro  Ctrl+R', view.toggle_macro_record),
                       ('P&lay Macro    Ctrl+P', view.play_macro),
                       ]),
            ('&Options', [('&Colors', view.on_colors),
                          ('&Editor', view.on_cfg_editor),
                          ]),
            ('&Help', [('&About', view.on_help_about),
                       ]),
            ]
    bar = Menu('')
    fill_menu(bar, desc)
    return bar


def main():
    app = Application()
    filename = ''
    if len(sys.argv) > 1:
        filename = sys.argv[1]
    doc = Document(filename)
    doc.load(filename)
    view = View(app, doc)
    app.set_menu(create_menu(view))
    app.draw_menu()
    view.render()
    try:
        while view.process_input():
            app.draw_menu()
            view.render()
    except utils.ExitException:
        pass

    app.close()


if __name__ == '__main__':
    main()



class ExitException(Exception):
    pass


def count_leading_spaces(s):
    n = 0
    for c in s:
        if c != ' ':
            break
        n += 1
    return n


def align(s, n):
    if len(s) > n:
        return s[0:n]
    if len(s) < n:
        return s + ' ' * (n - len(s))
    return s


def ctrl(key):
    return ord(key) - ord('A') + 1



import os
import atexit
from configparser import ConfigParser

home = os.environ['HOME']
cfgpath = os.path.join(home, '.yated.ini')
cfg = ConfigParser()
cfg.read(cfgpath)
if 'config' not in cfg:
    cfg['config'] = {}
section = cfg['config']


def get(name, default=''):
    if name not in section:
        section[name] = default
    return section.get(name)


def getint(name, default=0):
    return int(get(name, str(default)))


def getbool(name, default=False):
    return get(name, str(default)) != 'False'


def set(name, value):
    section[name] = str(value)


def save_cfg():
    with open(cfgpath, 'w') as configfile:
        cfg.write(configfile)


atexit.register(save_cfg)



import collections
import curses
from yated.ptypes import Point, Rect
from . import dialogs
from yated.clip import copy, paste
from . import utils
from . import config


class View:
    def __init__(self, app, doc):
        self.app = app
        self.rect = Rect(self.app.rect.tl + Point(0, 1), self.app.rect.br - Point(0, 2))
        self.doc = doc
        self.active_menu = None
        self.active_dialog = None
        if self.doc.tabs_replaced:
            self.active_dialog = dialogs.MessageBox('Tab were replaced by spaces (Esc to continue)')
        self.recording = False
        self.editrect = None
        self.macro = []
        self.last_find_action = None
        self.offset = Point(0, 0)
        self.cursor = Point(0, 0)
        self.rownum_width = 0
        self.selection = None
        self.lastx = 0
        self.tabsize = config.getint('tabsize', 4)
        self.insert = True
        self.shifted_moves = {'KEY_SLEFT': 'KEY_LEFT',
                              'KEY_SRIGHT': 'KEY_RIGHT',
                              'KEY_SF': 'KEY_DOWN',
                              'KEY_SR': 'KEY_UP',
                              'KEY_SPREVIOUS': 'KEY_PPAGE',
                              'KEY_SNEXT': 'KEY_NPAGE',
                              'KEY_SHOME': 'KEY_HOME',
                              'KEY_SEND': 'KEY_END',
                              'kRIT6': 'kRIT5',
                              'kLFT6': 'kLFT5'
                              }
        self.movement_keys = {'KEY_LEFT': (-1, 0),
                              'KEY_RIGHT': (1, 0),
                              'KEY_DOWN': (0, 1),
                              'KEY_UP': (0, -1),
                              'KEY_PPAGE': (0, -self.rect.height()),
                              'KEY_NPAGE': (0, self.rect.height()),
                              'KEY_HOME': lambda v: (-v.cursor.x, 0),
                              'KEY_END': (99999, 0),
                              'kRIT5': lambda v: v.doc.word_right(v.cursor) - v.cursor,
                              'kLFT5': lambda v: v.doc.word_left(v.cursor) - v.cursor,
                              'kEND5': lambda v: (0, v.doc.rows_count()),
                              'kHOM5': lambda v: (-v.cursor.x, -v.cursor.y)
                              }

    def process_text_input(self, c):
        movement = None
        # if c == 27:
        #    raise utils.ExitException()
        if c == utils.ctrl('C'):
            self.on_copy()
        if c == utils.ctrl('F'):
            self.on_find_replace()
        if c == utils.ctrl('X'):
            self.on_cut()
        if c == utils.ctrl('V'):
            text = paste()
            if len(text) > 0:
                self.doc.start_compound()
                if self.selection is not None:
                    self.delete_selection()
                movement = self.doc.add_text(text, self.cursor, self.insert)
                self.doc.stop_compound()
        if c == utils.ctrl('N'):
            self.on_file_new()
        if c == utils.ctrl('S'):
            self.on_file_save()
        if c == utils.ctrl('O'):
            self.on_file_open()
        if c == utils.ctrl('Q'):
            self.on_file_exit()
        if c == utils.ctrl('R'):
            self.toggle_macro_record()
        if c == utils.ctrl('P'):
            self.play_macro()
        if c == 26:  # Ctrl+Z
            movement = self.doc.undo()
        if c == ord('/') and self.selection is not None:
            sel = self.normalized_selection()
            inc = 0
            if sel.br.x > 0:
                inc = 1
            self.doc.start_compound()
            for y in range(sel.tl.y, sel.br.y + inc):
                row = self.doc.get_row(y)
                if row.startswith('//'):
                    self.doc.delete_block(y, 0, 2)
                else:
                    self.doc.add_text('//', Point(0, y), True)
            self.doc.stop_compound()
        elif 32 <= c < 127:
            if self.selection is not None:
                self.delete_selection()
            if self.doc.add_char(chr(c), self.cursor, self.insert):
                movement = (1, 0)
        if c == 9:
            self.doc.start_compound()
            if self.selection is not None:
                sel = self.normalized_selection()
                inc = 0
                if sel.br.x > 0:
                    inc = 1
                for y in range(sel.tl.y, sel.br.y + inc):
                    self.doc.add_text(' ' * self.tabsize, Point(0, y), self.insert)
            else:
                movement = self.doc.add_text(' ' * self.tabsize, self.cursor, self.insert)
            self.doc.stop_compound()
        if c == 10:
            if self.selection is not None:
                self.delete_selection()
            row = self.doc.get_row(self.cursor.y)
            indent = 0
            for i in range(0, len(row)):
                if row[i] == ' ':
                    indent += 1
                else:
                    break
            movement = self.doc.new_line(self.cursor)
            self.doc.insert_block(' ' * indent, self.cursor + movement)
            movement += Point(indent, 0)
        return movement

    def process_movement(self, movement):
        movement = Point(movement)
        new_cursor = self.doc.set_cursor(self.cursor + movement)
        if movement.x != 0:
            self.lastx = new_cursor.x
        else:
            new_cursor = self.doc.set_cursor(Point(self.lastx, new_cursor.y))
        if self.selection is not None:
            self.selection.br = new_cursor
        self.cursor = new_cursor
        self.scroll_display()

    def scroll_display(self):
        scr_pos = self.cursor - self.offset + Point(self.rownum_width, 0)
        if not self.editrect.is_point_inside(scr_pos):
            if scr_pos.x >= self.rect.br.x:
                self.offset.x = self.cursor.x - self.editrect.width() + 1
            if scr_pos.x < self.editrect.tl.x:
                self.offset.x = self.cursor.x
            if scr_pos.y >= self.editrect.br.y or scr_pos.y < self.editrect.tl.y:
                self.offset.y = self.cursor.y - int(self.editrect.height() / 2)
                if self.offset.y < 0:
                    self.offset.y = 0
            self.doc.invalidate()

    def process_movement_key(self, key):
        movement = None
        shift = False
        if key in self.shifted_moves:
            if self.selection is None:
                self.selection = Rect(self.cursor, self.cursor)
            key = self.shifted_moves.get(key)
            shift = True
        if key in self.movement_keys:
            m = self.movement_keys.get(key)
            if hasattr(m,'__call__'):
                movement = m(self)
            else:
                movement = m
            if not shift:
                self.selection = None
        return movement

    def process_special_keys(self, key):
        movement = None
        if key == chr(0x7f):
            key = 'KEY_BACKSPACE'
        if key == 'KEY_F(3)':
            if self.last_find_action is not None:
                self.last_find_action()
        if key == 'KEY_DC':
            if self.selection is not None:
                self.delete_selection()
            else:
                self.doc.delete_char(self.cursor)
        if key == 'KEY_BACKSPACE':
            if self.selection is not None:
                self.delete_selection()
            else:
                if self.cursor.x > 0:
                    self.doc.delete_char(self.cursor - Point(1, 0))
                    movement = (-1, 0)
                elif self.cursor.y > 0:
                    y = self.cursor.y - 1
                    movement = (len(self.doc.get_row(y)) - self.cursor.x, -1)
                    self.doc.join_next_row(self.cursor.y - 1)
        if key == 'KEY_BTAB' and self.selection is not None:
            sel = self.normalized_selection()
            inc = 0
            if sel.br.x > 0:
                inc = 1
            for y in range(sel.tl.y, sel.br.y + inc):
                row = self.doc.get_row(y)
                n = utils.count_leading_spaces(row)
                if n > self.tabsize:
                    n = self.tabsize
                if n > 0:
                    self.doc.delete_block(y, 0, n)
        return movement

    def delete_selection(self):
        if self.selection is not None:
            self.doc.start_compound()
            sel = self.normalized_selection()
            if sel.tl.y == sel.br.y:
                self.doc.delete_block(sel.tl.y, sel.tl.x, sel.br.x)
            else:
                self.doc.delete_block(sel.tl.y, sel.tl.x, -1)
                self.doc.delete_block(sel.br.y, 0, sel.br.x)
                for y in range(sel.tl.y + 1, sel.br.y):
                    self.doc.delete_row(sel.tl.y + 1)
                self.doc.join_next_row(sel.tl.y)
            self.cursor = sel.tl
            self.doc.stop_compound()
        self.selection = None

    def process_app_shortcuts(self, key):
        if key in self.app.shortcuts:
            self.active_menu = self.app.shortcuts.get(key)

    def process_menu_keys(self, key):
        menu_index = self.app.menu_bar.items.index(self.active_menu)
        n = len(self.app.menu_bar.items)
        if key == 'KEY_LEFT':
            self.active_menu = self.app.menu_bar.items[(menu_index - 1) % n]
        if key == 'KEY_RIGHT':
            self.active_menu = self.app.menu_bar.items[(menu_index + 1) % n]
        if key == 'KEY_DOWN':
            self.active_menu.select_next()
        if key == 'KEY_UP':
            self.active_menu.select_prev()
        if self.active_menu.on_key(key):
            return
        if key == chr(10):
            self.active_menu.activate_current()
            self.active_menu = None
            return
        if key == 'ESC':
            self.active_menu = None
            return
        for item in self.active_menu.items:
            if key.lower() == item.key.lower():
                item.activate()
                self.active_menu = None
                return

    def process_dialog_keys(self, key):
        if key == 'ESC':
            self.active_dialog = None
        else:
            actions = self.active_dialog.process_key(key)
            if actions is not None:
                last_dialog = self.active_dialog
                if not isinstance(actions, list):
                    actions = [actions]
                for action in actions:
                    if hasattr(action,'__call__'):
                        action()
                if last_dialog == self.active_dialog:
                    self.active_dialog = None

    def process_input(self):
        key = self.app.getkey()
        if key == 'KEY_F(12)':
            raise utils.ExitException()
        if self.recording and key != chr(utils.ctrl('R')):
            self.macro.append(key)
        return self.process_key(key)

    def process_key(self, key):
        self.process_app_shortcuts(key)
        if self.active_menu is not None:
            self.process_menu_keys(key)
        elif self.active_dialog is not None:
            self.process_dialog_keys(key)
        else:
            movement = self.process_movement_key(key)
            if not movement:
                movement = self.process_special_keys(key)
            if not movement and len(key) == 1:
                movement = self.process_text_input(ord(key[0]))
            if movement:
                self.process_movement(movement)
        return True

    def normalized_selection(self):
        if self.selection is None:
            return None
        sel = self.selection
        if sel.tl.y > sel.br.y:
            sel = Rect(sel.br, sel.tl)
        if sel.tl.y == sel.br.y and sel.tl.x > sel.br.x:
            sel = Rect(sel.br, sel.tl)
        return sel

    def line_number_width(self):
        s = str(len(self.doc.text))
        return len(s) + 1

    def render(self):
        if not self.doc.valid:
            x0 = self.line_number_width()
            self.rownum_width = x0
            self.editrect = Rect(self.app.rect.tl + Point(x0, 1), self.app.rect.br - Point(0, 2))
            w = self.app.width - x0
            sel = self.normalized_selection()
            i0 = self.offset.y - 1
            j0 = self.offset.x
            for y in range(1, self.app.height - 1):
                row_index = i0 + y
                self.app.move(Point(0, y))
                rownum = str(row_index + 1)
                rownum = ' ' * (x0 - len(rownum) - 1) + rownum + ' '
                if row_index >= self.doc.rows_count():
                    rownum = ' ' * len(rownum)
                self.app.write(rownum, 2)
                self.app.move(Point(x0, y))
                row = ' ' * w
                if 0 <= row_index < self.doc.rows_count():
                    row = str(self.doc.get_row(row_index))
                    row = row[j0:]
                    if len(row) > w:
                        row = row[0:w]
                    if len(row) < w:
                        row = row + ' ' * (w - len(row))
                    if sel is not None:
                        if sel.tl.y <= row_index <= sel.br.y:
                            x = 0
                            if row_index == sel.tl.y:
                                x = sel.tl.x - j0
                                self.app.write(row[0:x], 1)
                            limit = len(row)
                            if row_index == sel.br.y:
                                limit = sel.br.x - j0
                            self.app.write(row[x:limit], 3)
                            if limit < len(row):
                                self.app.write(row[limit:], 1)
                        else:
                            self.app.write(row, 1)
                    else:
                        self.app.write(row, 1)
                else:
                    self.app.write(row, 2)
        self.app.move(Point(0, self.app.height - 1))
        s = utils.align(self.doc.filename, self.app.width - 1)
        self.app.write(s, 4)
        if self.recording:
            self.app.move((self.app.width - 1, 0))
            self.app.write('M', 7, 0)
        self.draw_cursor()
        if self.active_menu is not None:
            self.active_menu.draw(self.app)
        if self.active_dialog is not None:
            self.active_dialog.draw(self.app)
        self.app.refresh()

    def get_selected_text(self):
        sel = self.normalized_selection()
        res = []
        for row_index in range(sel.tl.y, sel.br.y + 1):
            row = self.doc.get_row(row_index)
            x = 0
            limit = len(row)
            if row_index == sel.tl.y:
                x = sel.tl.x
            if row_index == sel.br.y:
                limit = sel.br.x
            res.append(row[x:limit])
        return '\n'.join(res)

    def draw_cursor(self):
        p = self.cursor - self.offset + Point(self.rownum_width, 1)
        self.app.move(p)

    def on_copy(self):
        if self.selection is not None:
            copy(self.get_selected_text())
            return True
        return False

    def on_cut(self):
        if self.on_copy():
            self.delete_selection()

    def on_paste(self):
        curses.ungetch(22)

    def find(self, details):
        try:
            first = details.get('first')
            details['first'] = False
            self.last_find_action = lambda: self.find(details)
            text = details.get('find')
            case = details.get('case')
            regex = details.get('regex')
            word = details.get('word')
            config.set('find_text', text)
            config.set('find_case', case)
            config.set('find_regex', regex)
            config.set('find_word', word)
            count = 0
            start_cursor = Point(self.cursor)
            if not first:
                self.cursor = Point(self.cursor.x + 1, self.cursor.y)
            while count <= self.doc.rows_count():
                x = self.doc.find_in_row(self.cursor, text, case, regex, word)
                if x >= 0:
                    self.cursor = Point(x, self.cursor.y)
                    self.scroll_display()
                    return True
                else:
                    self.cursor = Point(0, (self.cursor.y + 1) % self.doc.rows_count())
                    count += 1
            self.cursor = start_cursor
            self.app.move(start_cursor)
        except KeyError:
            pass
        return False

    def replace(self, details):
        replace_all = details.get('all')
        text = details.get('find')
        new_text = details.get('replace')
        config.set('find_replace', new_text)
        while self.find(details):
            x = self.cursor.x
            self.doc.delete_block(self.cursor.y, x, x + len(text))
            self.doc.insert_block(new_text, self.cursor)
            self.last_find_action = lambda: self.replace(details)
            if not replace_all:
                break

    def on_find_replace(self):
        self.active_dialog = dialogs.FindReplaceDialog(self.find, self.replace)

    def on_find_again(self):
        if self.last_find_action is not None:
            self.last_find_action()

    def on_ask_save(self, action):
        if self.doc.modified:
            self.active_dialog = dialogs.MessageBox('Save File? Y/N')
            self.active_dialog.add_key('Y', [self.on_file_save, action])
            self.active_dialog.add_key('N', [self.on_file_discard, action])
            return True
        return False

    def on_file_new(self):
        if not self.on_ask_save(self.on_file_new):
            self.doc.clear()
            self.cursor = Point(0, 0)

    def on_file_load(self, filename):
        self.doc.load(filename)
        if self.doc.tabs_replaced:
            self.active_dialog = dialogs.MessageBox('Tab were replaced by spaces (Esc to continue)')

    def on_file_open(self):
        if not self.on_ask_save(self.on_file_open):
            self.active_dialog = dialogs.FileDialog(self.on_file_load)

    def on_file_save(self):
        if not self.doc.save():
            self.active_dialog = dialogs.FileDialog(self.doc.save_as)

    def on_file_save_as(self):
        self.active_dialog = dialogs.FileDialog(self.doc.save_as)

    def on_file_exit(self):
        if not self.on_ask_save(self.on_file_exit):
            raise utils.ExitException()

    def on_file_discard(self):
        self.doc.modified = False

    def toggle_macro_record(self):
        if not self.recording:
            self.macro = []
        self.recording = not self.recording

    def play_macro(self):
        if self.recording:
            self.recording = False
        for key in self.macro:
            self.process_key(key)

    def on_colors(self):
        self.active_dialog = dialogs.ColorConfigDialog()

    def update_editor_options(self, details):
        self.tabsize = int(details.get('tabsize'))
        config.set('tabsize', self.tabsize)

    def on_cfg_editor(self):
        self.active_dialog = dialogs.EditorOptionsDialog(self.update_editor_options)

    def on_help_about(self):
        self.active_dialog = dialogs.AboutDialog()
